#!/usr/bin/env python2.7

### Cesar Rodriguez Sept 2022
### changes pload mass depending on /status/pload_mass parameter

import rospy
import rosservice
from gazebo_msgs.srv import GetLinkProperties, SetLinkProperties

class Main:
    def __init__(self):
        # Variables needed for testing start

        # link (payload) name
        self.link_name =  'spiri_with_tether::mass::payload'

        # service names we will use
        self.service_request = '/gazebo/get_link_properties'
        self.service_call = '/gazebo/set_link_properties'

        # wait for service
        rospy.wait_for_service(self.service_request)
        # set up service request
        get_link_properties = rospy.ServiceProxy(self.service_request,GetLinkProperties)
        # wait for service to set link properties
        rospy.wait_for_service(self.service_call)
        # set up service request
        set_link_properties = rospy.ServiceProxy(self.service_call,SetLinkProperties)

        # get payload properties
        self.pload_properties = get_link_properties(self.link_name)

        # state parameter name for where payload mass is stored
        self.param_pload_mass = 'status/pload_mass'
        
        # check if parameter exists
        if rospy.has_param(self.param_pload_mass):
            self.desired_mass = rospy.get_param(self.param_pload_mass) # get desired mass
            try:
                # set desired mass and necessary inertial properties
                set_link_properties(link_name=self.link_name,mass=self.desired_mass,
                    ixx=self.pload_properties.ixx,iyy=self.pload_properties.iyy,izz=self.pload_properties.izz,
                    gravity_mode=self.pload_properties.gravity_mode) 
                rospy.loginfo('Set payload mass to: %.3f', self.desired_mass)
            except rospy.ServiceException as e:
                rospy.loginfo("Set Link State call failed: {0}".format(e))
        else: rospy.logwarn('COULD NOT FIND PLOAD MASS PARAM')
			
if __name__=="__main__":

	# Initiate ROS node
	rospy.init_node('MoCap_node',anonymous=False)
	try:
		Main() # create class object
		rospy.spin() # loop until shutdown signal

	except rospy.ROSInterruptException:
	        pass

